/*
 *		ZpX^bN
 *
 *		T.Kobayashi		1993.8.6
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "data.h"
#include "alloc.h"

#include "inlib.h"
#include "err.h"

extern	int		ExecDebug ;

/*	ZpX^bN	*/
int			MaxStack ;
DataStruct	*StackBuf ;
int			StackPtr ;
int			StackUsed ;

/*	X^bNgbv𓾂	*/
DataStruct	*StackTop()
{
	assert( -1 <= StackPtr && StackPtr < MaxStack );
	return (DataStruct*)&StackBuf[StackPtr] ;
}

/*	|bv	*/
void	StackPop()
{
	assert( 0 <= StackPtr && StackPtr < MaxStack );

	if ( StackBuf[StackPtr].type & TYPE_OBJECT )
	{
		ObjectFree( StackBuf[StackPtr].od.ptr );
	}
	StackPtr -- ;
#if	PDEBUG
	if ( ExecDebug )
	{
		printf( "  " );
		printf( "POP [%d] ", StackPtr );
		DataPrint( &StackBuf[StackPtr+1] );
		printf( "\n" );
	}
#endif
}

/*	vbV	*/
void	StackPush( buf )
DataStruct	*buf ;
{
	StackPtr ++ ;
	assert( 0 <= StackPtr );
	if ( StackUsed < StackPtr )
		StackUsed = StackPtr ;
	if ( StackPtr >= MaxStack )
	{
		ExecError( "X^bNI[o[t[II" );
	}
	if ( buf == NULL )
	{
		buf->type = TYPE_NOASN ;
	}
	else
	{
		StackBuf[StackPtr] = *buf ;
		if ( buf->type & TYPE_OBJECT )
		{
			StackBuf[StackPtr].od.ptr = ObjectCopy( buf->od.ptr );
		}
	}

#if	PDEBUG
	if ( ExecDebug )
	{
		printf( "  " );
		printf( "PUSH[%d] ", StackPtr );
		DataPrint( buf );
		printf( "\n" );
	}
#endif

}

/*	_^vbV	*/
void	StackPushBoolean( l )
int		l ;
{
	StackPtr ++ ;
	assert( 0 <= StackPtr );
	if ( StackUsed < StackPtr )
		StackUsed = StackPtr ;
	if ( StackPtr >= MaxStack )
	{
		ExecError( "X^bNI[o[t[II" );
	}
	StackBuf[StackPtr].type = TYPE_BOOLEAN ;
	StackBuf[StackPtr].ld.l = l ;

#if	PDEBUG
	if ( ExecDebug )
	{
		printf( "  " );
		printf( "PUSH[%d] int : %d\n", StackPtr, l );
	}
#endif
}


/*	vbV	*/
void	StackPushInt( i )
int		i ;
{
	StackPtr ++ ;
	assert( 0 <= StackPtr );
	if ( StackUsed < StackPtr )
		StackUsed = StackPtr ;
	if ( StackPtr >= MaxStack )
	{
		ExecError( "X^bNI[o[t[II" );
	}
	StackBuf[StackPtr].type = TYPE_INT ;
	StackBuf[StackPtr].id.i = i ;

#if	PDEBUG
	if ( ExecDebug )
	{
		printf( "  " );
		printf( "PUSH[%d] int : %d\n", StackPtr, i );
	}
#endif
}

/*	vbV	*/
void	StackPushReal( r )
double	r ;
{
	StackPtr ++ ;
	assert( 0 <= StackPtr );
	if ( StackUsed < StackPtr )
		StackUsed = StackPtr ;
	if ( StackPtr >= MaxStack )
	{
		ExecError( "X^bNI[o[t[II" );
	}
	StackBuf[StackPtr].type = TYPE_REAL ;
	StackBuf[StackPtr].rd.r = (float)r ;

#if	PDEBUG
	if ( ExecDebug )
	{
		printf( "  " );
		printf( "PUSH[%d] int : %d\n", StackPtr, r );
	}
#endif
}

/*	X^bNɗ̈𓾂	*/
DataStruct	*StackAlloc( size )
int		size ;
{
	int		i ;
	DataStruct	*data, *ret ;

	assert( 0 <= size );

	data = & StackBuf[StackPtr] + 1;
	ret = data ;
	StackPtr += size ;

	assert( 0 <= StackPtr );
	if ( StackUsed < StackPtr )
		StackUsed = StackPtr ;
	if ( StackPtr >= MaxStack )
	{
		ExecError( "X^bNI[o[t[II" );
	}

	for( i = 0 ; i < size ; i++ )
	{
		data->type = TYPE_NOASN ;
		data++ ;
	}

#if	PDEBUG
	if ( ExecDebug )
	{
		printf( "  " );
		printf( "Alloc[%d] : %d\n", StackPtr, size );
	}
#endif

	return ret ;
}

/*	}[Nʒu܂ŉ	*/
void	StackRelease( ptr )
DataStruct	*ptr ;
{
	DataStruct	*top ;

	top = StackTop();
	assert( StackBuf - 1 <= ptr && ptr <= top );

	while( ptr < top )
	{
		if ( top->type & TYPE_OBJECT )
		{
			ObjectFree( top->od.ptr );
		}
		top -- ;
	}
	StackPtr = top - StackBuf ;

#if	PDEBUG
	if ( ExecDebug )
	{
		printf( "  " );
		printf( "Release[%d]\n", StackPtr );
	}
#endif

}

/*	X^bN̍őgpʂ𓾂	*/
int		StackMaxUsed()
{
	return StackUsed ;
}

/*	f[^obt@̓e\	*/
void	DataPrint( buf )
DataStruct	*buf ;
{
	char	str[256] ;

	DataToString( str, buf );
	printf( str );
}

/*	ɕϊ	*/
int		DataToString( str, buf )
char	*str ;
DataStruct	*buf ;
{
	char	*typename ;

	switch( buf->type )
	{
		case TYPE_BOOLEAN:
			if ( buf->ld.l )
				strcpy( str, "TRUE" );
			else
				strcpy( str, "FALSE" );
			break ;
		case TYPE_INT:
			sprintf( str, "%d", buf->id.i );
			break ;
		case TYPE_REAL:
			sprintf( str, "%f", buf->rd.r );
			break ;
		case TYPE_FUNC:
			sprintf( str, "%d", buf->funcd.ident );
			break ;
		case TYPE_ARRAY:
			sprintf( str, "%X[%d]", (int)(buf->ad.ary), buf->ad.size );
			break ;
		case TYPE_TYPE:
			switch( buf->td.t )
			{
				case TYPE_NOASN:
					typename = "NO_ASSIGN" ;
					break ;
				case TYPE_BOOLEAN:
					typename = "NO_ASSIGN" ;
				case TYPE_INT:
					typename = "INT" ;
					break ;
				case TYPE_REAL:
					typename = "REAL" ;
					break ;
				case TYPE_TYPE:
					typename = "TYPE" ;
					break ;
				case TYPE_ARRAY:
					typename = "ARRAY" ;
					break ;
				case TYPE_FUNC:
					typename = "FUNC" ;
					break ;
				case TYPE_OBJECT:
					typename = "OBJECT" ;
					break ;
				default:
					typename = "DATA TYPE ERROR!!!" ;
					break ;
			}
			strcpy( str, typename );
			break ;
		case TYPE_OBJECT:
			sprintf( str, "Object[%d][%X]", buf->od.ptr->type, buf->od.ptr );
			break ;
		default:
			assert( FALSE );
	}
	return strlen( str );
}

